"""GeoAlchemy2 package."""
import os

import sqlalchemy
from packaging import version
from sqlalchemy import Column
from sqlalchemy import Index
from sqlalchemy import Table
from sqlalchemy import event
from sqlalchemy import text
from sqlalchemy.sql import expression
from sqlalchemy.sql import func
from sqlalchemy.sql import select
from sqlalchemy.types import TypeDecorator

from . import functions  # noqa
from . import types  # noqa
from .elements import RasterElement  # noqa
from .elements import WKBElement  # noqa
from .elements import WKTElement  # noqa
from .exc import ArgumentError
from .types import Geography
from .types import Geometry
from .types import Raster
from .types import _DummyGeometry

_SQLALCHEMY_VERSION_BEFORE_14 = version.parse(sqlalchemy.__version__) < version.parse("1.4")


def _format_select_args(*args):
    if _SQLALCHEMY_VERSION_BEFORE_14:
        return [args]
    else:
        return args


def _check_spatial_type(tested_type, spatial_types, dialect=None):
    return (
        isinstance(tested_type, spatial_types)
        or (
            isinstance(tested_type, TypeDecorator)
            and isinstance(tested_type.load_dialect_impl(dialect), spatial_types)
        )
    )


def _spatial_idx_name(table_name, column_name):
    return 'idx_{}_{}'.format(table_name, column_name)


def check_management(column, dialect):
    return getattr(column.type, "management", False) is True or dialect.name == 'sqlite'


def _get_gis_cols(table, spatial_types, dialect, check_col_management=False):
    return [
        col
        for col in table.columns
        if (
            isinstance(col, Column)
            and _check_spatial_type(col.type, spatial_types, dialect)
            and (
                not check_col_management
                or check_management(col, dialect)
            )
        )
    ]


def _get_spatialite_attrs(bind, table_name, col_name):
    col_attributes = bind.execute(
        text("""SELECT * FROM "geometry_columns"
           WHERE f_table_name = '{}' and f_geometry_column = '{}'
        """.format(
            table_name, col_name
        ))
    ).fetchone()
    return col_attributes


def _get_spatialite_version(bind):
    return bind.execute(text("""SELECT spatialite_version();""")).fetchone()[0]


def _setup_ddl_event_listeners():
    @event.listens_for(Table, "before_create")
    def before_create(target, connection, **kw):
        dispatch("before-create", target, connection)

    @event.listens_for(Table, "after_create")
    def after_create(target, connection, **kw):
        dispatch("after-create", target, connection)

    @event.listens_for(Table, "before_drop")
    def before_drop(target, connection, **kw):
        dispatch("before-drop", target, connection)

    @event.listens_for(Table, "after_drop")
    def after_drop(target, connection, **kw):
        dispatch("after-drop", target, connection)

    @event.listens_for(Column, 'after_parent_attach')
    def after_parent_attach(column, table):
        if not isinstance(table, Table):
            # For old versions of SQLAlchemy, subqueries might trigger the after_parent_attach event
            # with a selectable as table, so we want to skip this case.
            return

        if (
            not getattr(column.type, "spatial_index", False)
            and getattr(column.type, "use_N_D_index", False)
        ):
            raise ArgumentError('Arg Error(use_N_D_index): spatial_index must be True')

        if (
            getattr(column.type, "management", True)
            or not getattr(column.type, "spatial_index", False)
        ):
            # If the column is managed, the indexes are created after the table
            return

        try:
            if column.type._spatial_index_reflected:
                return
        except AttributeError:
            pass

        if _check_spatial_type(column.type, (Geometry, Geography)):
            if column.type.use_N_D_index:
                postgresql_ops = {column.name: "gist_geometry_ops_nd"}
            else:
                postgresql_ops = {}
            table.append_constraint(
                Index(
                    _spatial_idx_name(table.name, column.name),
                    column,
                    postgresql_using='gist',
                    postgresql_ops=postgresql_ops,
                    _column_flag=True,
                )
            )
        elif _check_spatial_type(column.type, Raster):
            table.append_constraint(
                Index(
                    _spatial_idx_name(table.name, column.name),
                    func.ST_ConvexHull(column),
                    postgresql_using='gist',
                    _column_flag=True,
                )
            )

    def dispatch(current_event, table, bind):
        if current_event in ('before-create', 'before-drop'):
            # Filter Geometry columns from the table with management=True
            # Note: Geography and PostGIS >= 2.0 don't need this
            gis_cols = _get_gis_cols(table, Geometry, bind.dialect, check_col_management=True)

            # Find all other columns that are not managed Geometries
            regular_cols = [x for x in table.columns if x not in gis_cols]

            # Save original table column list for later
            table.info["_saved_columns"] = table.columns

            # Temporarily patch a set of columns not including the
            # managed Geometry columns
            column_collection = expression.ColumnCollection()
            for col in regular_cols:
                column_collection.add(col)
            table.columns = column_collection

            if current_event == 'before-drop':
                # Drop the managed Geometry columns
                for col in gis_cols:
                    if bind.dialect.name == 'sqlite':
                        drop_func = 'DiscardGeometryColumn'

                        # Disable spatial indexes if present
                        stmt = select(
                            *_format_select_args(
                                getattr(func, 'CheckSpatialIndex')(table.name, col.name)
                            )
                        )
                        if bind.execute(stmt).fetchone()[0] is not None:
                            stmt = select(
                                *_format_select_args(
                                    getattr(func, 'DisableSpatialIndex')(table.name, col.name)
                                )
                            )
                            stmt = stmt.execution_options(autocommit=True)
                            bind.execute(stmt)
                            bind.execute(
                                text(
                                    """DROP TABLE IF EXISTS {};""".format(
                                        _spatial_idx_name(
                                            table.name,
                                            col.name,
                                        )
                                    )
                                )
                            )
                    elif bind.dialect.name == 'postgresql':
                        drop_func = 'DropGeometryColumn'
                    else:
                        raise ArgumentError('dialect {} is not supported'.format(bind.dialect.name))
                    args = [table.schema] if table.schema else []
                    args.extend([table.name, col.name])

                    stmt = select(*_format_select_args(getattr(func, drop_func)(*args)))
                    stmt = stmt.execution_options(autocommit=True)
                    bind.execute(stmt)
            elif current_event == 'before-create':
                # Remove the spatial indexes from the table metadata because they should not be
                # created during the table.create() step since the associated columns do not exist
                # at this time.
                table.info["_after_create_indexes"] = []
                current_indexes = set(table.indexes)
                for idx in current_indexes:
                    for col in table.info["_saved_columns"]:
                        if (
                            _check_spatial_type(col.type, Geometry, bind.dialect)
                            and check_management(col, bind.dialect)
                        ) and col in idx.columns.values():
                            table.indexes.remove(idx)
                            if (
                                idx.name != _spatial_idx_name(table.name, col.name)
                                or not getattr(col.type, "spatial_index", False)
                            ):
                                table.info["_after_create_indexes"].append(idx)
                if bind.dialect.name == 'sqlite':
                    for col in gis_cols:
                        # Add dummy columns with GEOMETRY type
                        col._actual_type = col.type
                        col.type = _DummyGeometry()
                        col.nullable = col._actual_type.nullable
                    table.columns = table.info["_saved_columns"]

        elif current_event == 'after-create':
            # Restore original column list including managed Geometry columns
            table.columns = table.info.pop('_saved_columns')

            for col in table.columns:
                # Add the managed Geometry columns with AddGeometryColumn()
                if (
                    _check_spatial_type(col.type, Geometry, bind.dialect)
                    and check_management(col, bind.dialect)
                ):
                    dimension = col.type.dimension
                    if bind.dialect.name == 'sqlite':
                        col.type = col._actual_type
                        del col._actual_type
                        create_func = func.RecoverGeometryColumn
                        if col.type.dimension == 4:
                            dimension = 'XYZM'
                        elif col.type.dimension == 2:
                            dimension = 'XY'
                        else:
                            if col.type.geometry_type.endswith('M'):
                                dimension = 'XYM'
                            else:
                                dimension = 'XYZ'
                    else:
                        create_func = func.AddGeometryColumn
                    args = [table.schema] if table.schema else []
                    args.extend([
                        table.name,
                        col.name,
                        col.type.srid,
                        col.type.geometry_type,
                        dimension
                    ])
                    if col.type.use_typmod is not None:
                        args.append(col.type.use_typmod)

                    stmt = select(*_format_select_args(create_func(*args)))
                    stmt = stmt.execution_options(autocommit=True)
                    bind.execute(stmt)

                # Add spatial indices for the Geometry and Geography columns
                if (
                    _check_spatial_type(col.type, (Geometry, Geography), bind.dialect)
                    and col.type.spatial_index is True
                ):
                    if bind.dialect.name == 'sqlite':
                        stmt = select(*_format_select_args(func.CreateSpatialIndex(table.name,
                                                                                   col.name)))
                        stmt = stmt.execution_options(autocommit=True)
                        bind.execute(stmt)
                    elif bind.dialect.name == 'postgresql':
                        # If the index does not exist (which might be the case when
                        # management=False), define it and create it
                        if (
                            not [i for i in table.indexes if col in i.columns.values()]
                            and check_management(col, bind.dialect)
                        ):
                            if col.type.use_N_D_index:
                                postgresql_ops = {col.name: "gist_geometry_ops_nd"}
                            else:
                                postgresql_ops = {}
                            idx = Index(
                                _spatial_idx_name(table.name, col.name),
                                col,
                                postgresql_using='gist',
                                postgresql_ops=postgresql_ops,
                                _column_flag=True,
                            )
                            idx.create(bind=bind)

                    else:
                        raise ArgumentError('dialect {} is not supported'.format(bind.dialect.name))

                if (
                    isinstance(col.type, (Geometry, Geography))
                    and col.type.spatial_index is False
                    and col.type.use_N_D_index is True
                ):
                    raise ArgumentError('Arg Error(use_N_D_index): spatial_index must be True')

            for idx in table.info.pop("_after_create_indexes"):
                table.indexes.add(idx)
                idx.create(bind=bind)

        elif current_event == 'after-drop':
            # Restore original column list including managed Geometry columns
            table.columns = table.info.pop('_saved_columns')

    @event.listens_for(Table, "column_reflect")
    def _reflect_geometry_column(inspector, table, column_info):
        if not isinstance(column_info.get("type"), Geometry):
            return

        if inspector.bind.dialect.name == "postgresql":
            geo_type = column_info["type"]
            geometry_type = geo_type.geometry_type
            coord_dimension = geo_type.dimension
            if geometry_type.endswith("ZM"):
                coord_dimension = 4
            elif geometry_type[-1] in ["Z", "M"]:
                coord_dimension = 3

            # Query to check a given column has spatial index
            if table.schema is not None:
                schema_part = " AND nspname = '{}'".format(table.schema)
            else:
                schema_part = ""

            has_index_query = """SELECT (indexrelid IS NOT NULL) AS has_index
                FROM (
                    SELECT
                            n.nspname,
                            c.relname,
                            c.oid AS relid,
                            a.attname,
                            a.attnum
                    FROM pg_attribute a
                    INNER JOIN pg_class c ON (a.attrelid=c.oid)
                    INNER JOIN pg_type t ON (a.atttypid=t.oid)
                    INNER JOIN pg_namespace n ON (c.relnamespace=n.oid)
                    WHERE t.typname='geometry'
                            AND c.relkind='r'
                ) g
                LEFT JOIN pg_index i ON (g.relid = i.indrelid AND g.attnum = ANY(i.indkey))
                WHERE relname = '{}' AND attname = '{}'{};
            """.format(
                table.name, column_info["name"], schema_part
            )
            spatial_index = inspector.bind.execute(text(has_index_query)).scalar()

            # Set attributes
            column_info["type"].geometry_type = geometry_type
            column_info["type"].dimension = coord_dimension
            column_info["type"].spatial_index = bool(spatial_index)

            # Spatial indexes are automatically reflected with PostgreSQL dialect
            column_info["type"]._spatial_index_reflected = True
        elif inspector.bind.dialect.name == "sqlite":
            # Get geometry type, SRID and spatial index from the SpatiaLite metadata
            col_attributes = _get_spatialite_attrs(inspector.bind, table.name, column_info["name"])
            if col_attributes is not None:
                _, _, geometry_type, coord_dimension, srid, spatial_index = col_attributes

                if isinstance(geometry_type, int):
                    geometry_type_str = str(geometry_type)
                    if geometry_type >= 1000:
                        first_digit = geometry_type_str[0]
                        has_z = first_digit in ["1", "3"]
                        has_m = first_digit in ["2", "3"]
                    else:
                        has_z = has_m = False
                    geometry_type = {
                        "0": "GEOMETRY",
                        "1": "POINT",
                        "2": "LINESTRING",
                        "3": "POLYGON",
                        "4": "MULTIPOINT",
                        "5": "MULTILINESTRING",
                        "6": "MULTIPOLYGON",
                        "7": "GEOMETRYCOLLECTION",
                    }[geometry_type_str[-1]]
                    if has_z:
                        geometry_type += "Z"
                    if has_m:
                        geometry_type += "M"
                else:
                    if "Z" in coord_dimension:
                        geometry_type += "Z"
                    if "M" in coord_dimension:
                        geometry_type += "M"
                    coord_dimension = {
                        "XY": 2,
                        "XYZ": 3,
                        "XYM": 3,
                        "XYZM": 4,
                    }.get(coord_dimension, coord_dimension)

                # Set attributes
                column_info["type"].geometry_type = geometry_type
                column_info["type"].dimension = coord_dimension
                column_info["type"].srid = srid
                column_info["type"].spatial_index = bool(spatial_index)

                # Spatial indexes are not automatically reflected with SQLite dialect
                column_info["type"]._spatial_index_reflected = False


_setup_ddl_event_listeners()


def load_spatialite(dbapi_conn, connection_record):
    """Load SpatiaLite extension in SQLite DB.

    The path to the SpatiaLite module should be set in the `SPATIALITE_LIBRARY_PATH` environment
    variable.
    """
    if "SPATIALITE_LIBRARY_PATH" not in os.environ:
        raise RuntimeError(
            "The SPATIALITE_LIBRARY_PATH environment variable is not set."
        )
    dbapi_conn.enable_load_extension(True)
    dbapi_conn.load_extension(os.environ["SPATIALITE_LIBRARY_PATH"])
    dbapi_conn.enable_load_extension(False)
    dbapi_conn.execute("SELECT InitSpatialMetaData();")


# Get version number
__version__ = "UNKNOWN VERSION"

# Attempt to use importlib.metadata first because it's much faster
# though it's only available in Python 3.8+ so we'll need to fall
# back to pkg_resources for Python 3.7 support
try:
    import importlib.metadata
except ImportError:
    try:
        from pkg_resources import DistributionNotFound
        from pkg_resources import get_distribution
    except ImportError:  # pragma: no cover
        pass
    else:
        try:
            __version__ = get_distribution('GeoAlchemy2').version
        except DistributionNotFound:  # pragma: no cover
            pass
else:
    try:
        __version__ = importlib.metadata.version('GeoAlchemy2')
    except importlib.metadata.PackageNotFoundError:  # pragma: no cover
        pass



"""
This module provides utility functions for integrating with Shapely.

.. note::

    As GeoAlchemy 2 itself has no dependency on `Shapely`, applications using
    functions of this module have to ensure that `Shapely` is available.
"""
import shapely.wkb
import shapely.wkt
from packaging import version

from .elements import WKBElement
from .elements import WKTElement

if version.parse(shapely.__version__) < version.parse('1.7'):  # pragma: no cover
    ######################################################################
    # Backport function from Shapely 1.7
    from shapely.geometry.base import geom_factory
    from shapely.geos import WKBWriter
    from shapely.geos import lgeos

    def dumps(ob, hex=False, srid=None, **kw):
        """Dump a WKB representation of a geometry to a byte string, or a
        hex-encoded string if ``hex=True``.

        Args:
            ob (geometry):
                The geometry to export to well-known binary (WKB) representation
            hex (bool):
                If true, export the WKB as a hexadecimal string. The default is to
                return a binary string/bytes object.
            srid (int):
                Spatial reference system ID to include in the output. The default
                value means no SRID is included.
        Keyword args:
            kwargs: See available keyword output settings in ``shapely.geos.WKBWriter``.
        """
        if srid is not None:
            # clone the object and set the SRID before dumping
            geom = lgeos.GEOSGeom_clone(ob._geom)
            lgeos.GEOSSetSRID(geom, srid)
            ob = geom_factory(geom)
            kw["include_srid"] = True
        writer = WKBWriter(lgeos, **kw)
        if hex:
            return writer.write_hex(ob)
        else:
            return writer.write(ob)
    ######################################################################
else:
    from shapely.wkb import dumps  # noqa


def to_shape(element):
    """
    Function to convert a :class:`geoalchemy2.types.SpatialElement`
    to a Shapely geometry.

    Args:
        element: The element to convert into a ``Shapely`` object.

    Example::

        lake = Session.query(Lake).get(1)
        polygon = to_shape(lake.geom)
    """
    assert isinstance(element, (WKBElement, WKTElement))
    if isinstance(element, WKBElement):
        data, hex = (element.data, True) if isinstance(element.data, str) else \
                    (bytes(element.data), False)
        return shapely.wkb.loads(data, hex=hex)
    elif isinstance(element, WKTElement):
        if element.extended:
            return shapely.wkt.loads(element.data.split(';', 1)[1])
        else:
            return shapely.wkt.loads(element.data)


def from_shape(shape, srid=-1, extended=False):
    """
    Function to convert a Shapely geometry to a
    :class:`geoalchemy2.types.WKBElement`.

    Args:
        srid:
            An integer representing the spatial reference system. E.g. ``4326``.
            Default value is ``-1``, which means no/unknown reference system.
        extended:
            A boolean to switch between WKB and EWKB.
            Default value is False.

    Example::

        from shapely.geometry import Point
        wkb_element = from_shape(Point(5, 45), srid=4326)
        ewkb_element = from_shape(Point(5, 45), srid=4326, extended=True)
    """
    return WKBElement(
        memoryview(dumps(shape, srid=srid if extended else None)),
        srid=srid,
        extended=extended)



"""
Compute length on insert
========================

It is possible to insert a geometry and ask PostgreSQL to compute its length at the same
time.
This example uses SQLAlchemy core queries.
"""
from sqlalchemy import Column
from sqlalchemy import Float
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import bindparam
from sqlalchemy import func

from geoalchemy2 import Geometry
from geoalchemy2.shape import to_shape

# Tests imports
from tests import select
from tests import test_only_with_dialects

metadata = MetaData()

table = Table(
    "inserts",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("geom", Geometry("LINESTRING", 4326)),
    Column("distance", Float),
)


@test_only_with_dialects("postgresql")
class TestLengthAtInsert():

    def test_query(self, conn):
        metadata.drop_all(conn, checkfirst=True)
        metadata.create_all(conn)

        # Define geometries to insert
        values = [
            {"ewkt": "SRID=4326;LINESTRING(0 0, 1 0)"},
            {"ewkt": "SRID=4326;LINESTRING(0 0, 0 1)"}
        ]

        # Define the query to compute distance (without spheroid)
        distance = func.ST_Length(func.ST_GeomFromText(bindparam("ewkt")), False)

        i = table.insert()
        i = i.values(geom=bindparam("ewkt"), distance=distance)

        # Execute the query with values as parameters
        conn.execute(i, values)

        # Check the result
        q = select([table])
        res = conn.execute(q).fetchall()

        # Check results
        assert len(res) == 2

        r1 = res[0]
        assert r1[0] == 1
        assert r1[1].srid == 4326
        assert to_shape(r1[1]).wkt == "LINESTRING (0 0, 1 0)"
        assert round(r1[2]) == 111195

        r2 = res[1]
        assert r2[0] == 2
        assert r2[1].srid == 4326
        assert to_shape(r2[1]).wkt == "LINESTRING (0 0, 0 1)"
        assert round(r2[2]) == 111195



import platform

import pytest
from shapely.geometry import LineString
from sqlalchemy import CheckConstraint
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy import text
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql import func

from geoalchemy2 import Geometry
from geoalchemy2.elements import WKTElement
from geoalchemy2.shape import from_shape

from . import select
from . import test_only_with_dialects

if platform.python_implementation().lower() == 'pypy':
    pytest.skip('skip SpatiaLite tests on PyPy', allow_module_level=True)


class TestIndex():
    @pytest.fixture
    def TableWithIndexes(self, base):
        class TableWithIndexes(base):
            __tablename__ = 'table_with_indexes'
            id = Column(Integer, primary_key=True)
            # Test indexes on Geometry columns.
            geom_not_managed_no_index = Column(
                Geometry(
                    geometry_type='POINT',
                    spatial_index=False,
                    management=False,
                )
            )
            geom_not_managed_index = Column(
                Geometry(
                    geometry_type='POINT',
                    spatial_index=True,
                    management=False,
                )
            )
            geom_managed_no_index = Column(
                Geometry(
                    geometry_type='POINT',
                    spatial_index=False,
                    management=True,
                )
            )
            geom_managed_index = Column(
                Geometry(
                    geometry_type='POINT',
                    spatial_index=True,
                    management=True,
                )
            )

        return TableWithIndexes

    @staticmethod
    def check_spatial_idx(bind, idx_name):
        tables = bind.execute(
            text("SELECT name FROM sqlite_master WHERE type ='table' AND name NOT LIKE 'sqlite_%';")
        ).fetchall()
        if idx_name in [i[0] for i in tables]:
            return True
        return False

    def test_index(self, conn, Lake, setup_tables):
        assert self.check_spatial_idx(conn, 'idx_lake_geom')

    def test_type_decorator_index(self, conn, LocalPoint, setup_tables):
        assert self.check_spatial_idx(conn, 'idx_local_point_geom')
        assert self.check_spatial_idx(conn, 'idx_local_point_managed_geom')

    def test_all_indexes(self, conn, TableWithIndexes, setup_tables):
        expected_indices = [
            'idx_table_with_indexes_geom_managed_index',
            'idx_table_with_indexes_geom_not_managed_index',
        ]
        for expected_idx in expected_indices:
            assert self.check_spatial_idx(conn, expected_idx)

        TableWithIndexes.__table__.drop(bind=conn)

        indexes_after_drop = conn.execute(text("""SELECT * FROM "geometry_columns";""")).fetchall()
        tables_after_drop = conn.execute(
            text("SELECT name FROM sqlite_master WHERE type ='table' AND name NOT LIKE 'sqlite_%';")
        ).fetchall()

        assert indexes_after_drop == []
        assert [table for table in tables_after_drop if 'table_with_indexes' in table.name] == []


class TestInsertionORM():
    pass


class TestUpdateORM():
    pass


class TestCallFunction():

    def test_ST_Buffer(self, session):
        """Test the specific SQLite signature with the `quadrantsegments` parameter."""
        s = select([
            func.St_AsText(func.ST_Buffer(WKTElement('LINESTRING(0 0,1 0)', srid=4326), 2, 1))
        ])
        r1 = session.execute(s).scalar()
        assert r1 == 'POLYGON((1 2, 3 0, 1 -2, 0 -2, -2 0, 0 2, 1 2))'

        s = select([
            func.St_AsText(func.ST_Buffer(WKTElement('LINESTRING(0 0,1 0)', srid=4326), 2, 2))
        ])
        r1 = session.execute(s).scalar()
        assert r1 == (
            'POLYGON((1 2, 2.414214 1.414214, 3 0, 2.414214 -1.414214, 1 -2, 0 -2, '
            '-1.414214 -1.414214, -2 0, -1.414214 1.414214, 0 2, 1 2))'
        )


class TestShapely():
    pass


class TestNullable():
    @pytest.fixture
    def NotNullableLake(self, base):
        class NotNullableLake(base):
            __tablename__ = 'NotNullablelake'
            id = Column(Integer, primary_key=True)
            geom = Column(Geometry(geometry_type='LINESTRING', srid=4326,
                                   management=True, nullable=False))

            def __init__(self, geom):
                self.geom = geom

        return NotNullableLake

    def test_insert(self, conn, NotNullableLake, setup_tables):
        # Insert geometries
        conn.execute(NotNullableLake.__table__.insert(), [
            {'geom': 'SRID=4326;LINESTRING(0 0,1 1)'},
            {'geom': WKTElement('LINESTRING(0 0,2 2)', srid=4326)},
            {'geom': from_shape(LineString([[0, 0], [3, 3]]), srid=4326)}
        ])

        # Fail when trying to insert null geometry
        with pytest.raises(IntegrityError):
            conn.execute(NotNullableLake.__table__.insert(), [
                {'geom': None}
            ])


class TestContraint():

    @pytest.fixture
    def ConstrainedLake(self, base):
        class ConstrainedLake(base):
            __tablename__ = 'contrained_lake'
            __table_args__ = (
                CheckConstraint(
                    '(geom is null and a_str is null) = (checked_str is null)', 'check_geom_sk'
                ),
            )
            id = Column(Integer, primary_key=True)
            a_str = Column(String, nullable=True)
            checked_str = Column(String, nullable=True)
            geom = Column(Geometry(geometry_type='LINESTRING', srid=4326, management=False))

            def __init__(self, geom):
                self.geom = geom

        return ConstrainedLake

    def test_insert(self, conn, ConstrainedLake, setup_tables):
        # Insert geometries
        conn.execute(ConstrainedLake.__table__.insert(), [
            {'a_str': None, 'geom': 'SRID=4326;LINESTRING(0 0,1 1)', 'checked_str': 'test'},
            {'a_str': 'test', 'geom': None, 'checked_str': 'test'},
            {'a_str': None, 'geom': None, 'checked_str': None},
        ])

        # Fail when trying to insert null geometry
        with pytest.raises(IntegrityError):
            conn.execute(ConstrainedLake.__table__.insert(), [
                {'a_str': None, 'geom': None, 'checked_str': 'should fail'},
            ])


class TestReflection():

    @pytest.fixture
    def setup_reflection_tables(self, reflection_tables_metadata, conn):
        reflection_tables_metadata.drop_all(conn, checkfirst=True)
        reflection_tables_metadata.create_all(conn)

    @test_only_with_dialects("sqlite-spatialite3")
    def test_reflection_spatialite_lt_4(self, conn, setup_reflection_tables):
        t = Table(
            'lake',
            MetaData(),
            autoload_with=conn)

        type_ = t.c.geom.type
        assert isinstance(type_, Geometry)
        assert type_.geometry_type == 'LINESTRING'
        assert type_.srid == 4326
        assert type_.dimension == 2

        type_ = t.c.geom_no_idx.type
        assert isinstance(type_, Geometry)
        assert type_.geometry_type == 'LINESTRING'
        assert type_.srid == 4326
        assert type_.dimension == 2

        type_ = t.c.geom_z.type
        assert isinstance(type_, Geometry)
        assert type_.geometry_type == 'LINESTRINGZ'
        assert type_.srid == 4326
        assert type_.dimension == 3

        type_ = t.c.geom_m.type
        assert isinstance(type_, Geometry)
        assert type_.geometry_type == 'LINESTRINGM'
        assert type_.srid == 4326
        assert type_.dimension == 3

        type_ = t.c.geom_zm.type
        assert isinstance(type_, Geometry)
        assert type_.geometry_type == 'LINESTRINGZM'
        assert type_.srid == 4326
        assert type_.dimension == 4

        # Drop the table
        t.drop(bind=conn)

        # Query to check the tables
        query_tables = text(
            """SELECT
                name
            FROM
                sqlite_master
            WHERE
                type ='table' AND
                name NOT LIKE 'sqlite_%'
            ORDER BY tbl_name;"""
        )

        # Query to check the indices
        query_indexes = text(
            """SELECT * FROM geometry_columns ORDER BY f_table_name, f_geometry_column;"""
        )

        # Check the indices
        geom_cols = conn.execute(query_indexes).fetchall()
        assert geom_cols == []

        # Check the tables
        all_tables = [i[0] for i in conn.execute(query_tables).fetchall()]
        assert all_tables == [
            'SpatialIndex',
            'geometry_columns',
            'geometry_columns_auth',
            'layer_statistics',
            'spatial_ref_sys',
            'spatialite_history',
            'views_geometry_columns',
            'views_layer_statistics',
            'virts_geometry_columns',
            'virts_layer_statistics',
        ]

        # Recreate the table to check that the reflected properties are correct
        t.create(bind=conn)

        # Check the actual properties
        geom_cols = conn.execute(query_indexes).fetchall()
        assert geom_cols == [
            ('lake', 'geom', 'LINESTRING', 'XY', 4326, 1),
            ('lake', 'geom_m', 'LINESTRING', 'XYM', 4326, 1),
            ('lake', 'geom_no_idx', 'LINESTRING', 'XY', 4326, 0),
            ('lake', 'geom_z', 'LINESTRING', 'XYZ', 4326, 1),
            ('lake', 'geom_zm', 'LINESTRING', 'XYZM', 4326, 1),
        ]

        all_tables = [i[0] for i in conn.execute(query_tables).fetchall()]
        assert all_tables == [
            'SpatialIndex',
            'geometry_columns',
            'geometry_columns_auth',
            'idx_lake_geom',
            'idx_lake_geom_m',
            'idx_lake_geom_m_node',
            'idx_lake_geom_m_parent',
            'idx_lake_geom_m_rowid',
            'idx_lake_geom_node',
            'idx_lake_geom_parent',
            'idx_lake_geom_rowid',
            'idx_lake_geom_z',
            'idx_lake_geom_z_node',
            'idx_lake_geom_z_parent',
            'idx_lake_geom_z_rowid',
            'idx_lake_geom_zm',
            'idx_lake_geom_zm_node',
            'idx_lake_geom_zm_parent',
            'idx_lake_geom_zm_rowid',
            'lake',
            'layer_statistics',
            'spatial_ref_sys',
            'spatialite_history',
            'views_geometry_columns',
            'views_layer_statistics',
            'virts_geometry_columns',
            'virts_layer_statistics',
        ]

    @test_only_with_dialects("sqlite-spatialite4")
    def test_reflection_spatialite_ge_4(self, conn, setup_reflection_tables):
        t = Table(
            'lake',
            MetaData(),
            autoload_with=conn)

        type_ = t.c.geom.type
        assert isinstance(type_, Geometry)
        assert type_.geometry_type == 'LINESTRING'
        assert type_.srid == 4326
        assert type_.dimension == 2

        type_ = t.c.geom_no_idx.type
        assert isinstance(type_, Geometry)
        assert type_.geometry_type == 'LINESTRING'
        assert type_.srid == 4326
        assert type_.dimension == 2

        type_ = t.c.geom_z.type
        assert isinstance(type_, Geometry)
        assert type_.geometry_type == 'LINESTRINGZ'
        assert type_.srid == 4326
        assert type_.dimension == 3

        type_ = t.c.geom_m.type
        assert isinstance(type_, Geometry)
        assert type_.geometry_type == 'LINESTRINGM'
        assert type_.srid == 4326
        assert type_.dimension == 3

        type_ = t.c.geom_zm.type
        assert isinstance(type_, Geometry)
        assert type_.geometry_type == 'LINESTRINGZM'
        assert type_.srid == 4326
        assert type_.dimension == 4

        # Drop the table
        t.drop(bind=conn)

        # Query to check the tables
        query_tables = text(
            """SELECT
                name
            FROM
                sqlite_master
            WHERE
                type ='table' AND
                name NOT LIKE 'sqlite_%'
            ORDER BY tbl_name;"""
        )

        # Query to check the indices
        query_indexes = text(
            """SELECT * FROM geometry_columns ORDER BY f_table_name, f_geometry_column;"""
        )

        # Check the indices
        geom_cols = conn.execute(query_indexes).fetchall()
        assert geom_cols == []

        # Check the tables
        all_tables = [i[0] for i in conn.execute(query_tables).fetchall()]
        assert all_tables == [
            'ElementaryGeometries',
            'SpatialIndex',
            'geometry_columns',
            'geometry_columns_auth',
            'geometry_columns_field_infos',
            'geometry_columns_statistics',
            'geometry_columns_time',
            'spatial_ref_sys',
            'spatial_ref_sys_aux',
            'spatialite_history',
            'sql_statements_log',
            'views_geometry_columns',
            'views_geometry_columns_auth',
            'views_geometry_columns_field_infos',
            'views_geometry_columns_statistics',
            'virts_geometry_columns',
            'virts_geometry_columns_auth',
            'virts_geometry_columns_field_infos',
            'virts_geometry_columns_statistics',
        ]

        # Recreate the table to check that the reflected properties are correct
        t.create(bind=conn)

        # Check the actual properties
        geom_cols = conn.execute(query_indexes).fetchall()
        assert geom_cols == [
            ('lake', 'geom', 2, 2, 4326, 1),
            ('lake', 'geom_m', 2002, 3, 4326, 1),
            ('lake', 'geom_no_idx', 2, 2, 4326, 0),
            ('lake', 'geom_z', 1002, 3, 4326, 1),
            ('lake', 'geom_zm', 3002, 4, 4326, 1),
        ]

        all_tables = [i[0] for i in conn.execute(query_tables).fetchall()]
        assert all_tables == [
            'ElementaryGeometries',
            'SpatialIndex',
            'geometry_columns',
            'geometry_columns_auth',
            'geometry_columns_field_infos',
            'geometry_columns_statistics',
            'geometry_columns_time',
            'idx_lake_geom',
            'idx_lake_geom_m',
            'idx_lake_geom_m_node',
            'idx_lake_geom_m_parent',
            'idx_lake_geom_m_rowid',
            'idx_lake_geom_node',
            'idx_lake_geom_parent',
            'idx_lake_geom_rowid',
            'idx_lake_geom_z',
            'idx_lake_geom_z_node',
            'idx_lake_geom_z_parent',
            'idx_lake_geom_z_rowid',
            'idx_lake_geom_zm',
            'idx_lake_geom_zm_node',
            'idx_lake_geom_zm_parent',
            'idx_lake_geom_zm_rowid',
            'lake',
            'spatial_ref_sys',
            'spatial_ref_sys_aux',
            'spatialite_history',
            'sql_statements_log',
            'views_geometry_columns',
            'views_geometry_columns_auth',
            'views_geometry_columns_field_infos',
            'views_geometry_columns_statistics',
            'virts_geometry_columns',
            'virts_geometry_columns_auth',
            'virts_geometry_columns_field_infos',
            'virts_geometry_columns_statistics',
        ]
